"""level.py

Part of the Tundra project.
(C) 2006 Rebuild Software
http://rebui.ld.net.au

This file contains the specification of Tundra's level class.
"""

import os, sys
import gzip
import xml.sax
import xml.sax.writer  # needed for data -> XML
from xml.sax.handler import ContentHandler

import globals as g
import style, player

try:
    import pygame
    from pygame.locals import *
except ImportError:
    raise ImportError, "You don't have Pygame installed."
try:
    assert pygame.image.get_extended()
except AssertionError:
    raise ImportError, "Your SDL_image does not support PNG images!"

ext = lambda s: s.split(".")[-1].lower()


#for a XML level file, if a level/object's style is classified as "internal",
#read the style data from the levelgroup's "custom" folder, xml has the same
#name as the level

class InteractObj:
    def __init__(self):
        self.posX = 0
        self.posY = 0
        self.frames = None
        self.objType = "harmless"
        self.activeRect = Rect((0,0),(16,16))
        self.drawFullGraphic = True
        self.drawIfTerrainUnderneath = False #tick
        self.drawButDontOverwrite = False  #tick, above flag doesn't work
        self.drawUpsideDown = False
        self.itemID = 0
        self.style = ""

class TerrainObj:
    def __init__(self):
        self.posX = 0
        self.posY = 0
        self.surface = None
        self.drawFullGraphic = True
        self.drawSubtractiveTerrain = False
        self.drawButDontOverwrite = False
        self.drawUpsideDown = False
        self.itemID = 0
        self.style = ""

class Level:
    """Level() -> Creates a new level context"""
    
    def __init__(self):
        self.releaseRate = 50
        self.numReleased = 10
        self.numToSave = 1
        self.timeLimit = 180
        self.numSkills = {"s_climber":10, "s_floater":10, 
                           "s_bomber":10, "s_blocker":10, 
                           "s_builder":10, "s_basher":10,
                           "s_miner":10, "s_digger":10 }
        self.startPosX = 400
        self.defaultStyle = "Soil"
        self.terrains = []
        self.interacts = []
        self.steels = []
        self.keyword = None

        self.levelName = "Another Lemming Caper"
        self.author = "Frosty LaHood Motors"
        self.version = "1.0"

        # self.levelBuf = pygame.Surface((320,240),HWSURFACE+SRCALPHA).convert()
        # order:
        # background + tiles + lemmings + mousecursor
    

    #THE COLLISION MAP:
    #(w, x, y, z) : 
    #w = (ground)
    #x = (exit, trap, water, hazard, leftonly, rightonly, steel)
    #y = (turnleft, turnright)
    #z = unused (always 255)


    def printCollisionMask(self):
        playfield = self.printTerrain()
        mask = pygame.Surface((playfield.get_width(), playfield.get_height()+2),SWSURFACE|SRCCOLORKEY)
        mask.fill((0,0,0,255))

        # arrrghghghg
        playfield.lock()
        mask.lock()
        for i in range(0, playfield.get_height()):
            for j in range(0, playfield.get_width()):
                w = 0
                if playfield.get_at((j,i))[3] == 255: #opaque, must be terrain!
                    w |= 128
                mask.fill( (w,0,0,255), ((j,i),(1,1)))

        for i in self.interacts:
            x = 0            
            if i.objType == "harmless":
                continue
            elif i.objType == "exit":
                x |= 128
            elif i.objType == "trap":
                x |= 64
            elif i.objType == "water":
                x |= 32
            elif i.objType == "hazard":
                x |= 16
            elif i.objType == "leftArrows":
                x |= 8
            elif i.objType == "rightArrows":
                x |= 4
            for j in range(i.posY + i.activeRect.top , i.posY + i.activeRect.bottom):
                for k in range(i.posX + i.activeRect.left , i.posX + i.activeRect.right):
                    try:   # "It Is Easier To Ask For Forgiveness Than To Ask For Permission"(TM)
                        ow, ox, oy, oz = mask.get_at((k,j))
                        mask.fill( (ow,x,oy,oz), ((k,j),(1,1)))
                    except IndexError:
                        continue

        for s in self.steels:
            x = 2
            for i in range(s.top , s.bottom):
                for j in range(s.left , s.right):
                    try:   # I love IIETAFFTTAFP(TM)
                        ow, ox, oy, oz = mask.get_at((j,i))
                        mask.fill( (ow,x|ox,oy,oz), ((j,i),(1,1)))
                    except IndexError:
                        continue

        mask.unlock()
        playfield.unlock()
#        pygame.image.save(mask, "test.tga")
        return mask

    def printTerrain(self, crop=False, showInter=False, preview=False):
        playfield = pygame.Surface([g.MAX_LEVEL_WIDTH,g.MAX_LEVEL_HEIGHT],SWSURFACE|SRCALPHA|SRCCOLORKEY).convert_alpha()
        playfield.fill((0,0,0,0))
    
        for t in self.terrains:
            if t.drawButDontOverwrite:
                blitBehind(t.surface, playfield, (t.posX, t.posY))
            else:
                if t.drawSubtractiveTerrain:
                    playfield.blit(t.subtractImg, (t.posX, t.posY))
                    playfield.set_colorkey(g.DEFAULT_COLOURKEY)
                    playfield = playfield.convert_alpha()
                    playfield.set_colorkey(g.DEFAULT_COLOURKEY2)
                    playfield = playfield.convert_alpha()
                else:
                    playfield.blit(t.surface, (t.posX, t.posY))

        if showInter:
            for i in self.interacts:
                if i.drawIfTerrainUnderneath:
                    scrub = i.frames[0].copy()
                    smackfu = playfield.get_rect().clip(((i.posX, i.posY),scrub.get_size()))
                    sampler = playfield.subsurface(smackfu)
                    scrub.lock()
                    sampler.lock()
                    for j in range(0,sampler.get_height()):
                        for k in range(0,sampler.get_width()):
                            if sampler.get_at((k,j))[3] == 0:
                                scrub.fill( (0,0,0,0), ((k,j),(1,1)))
                    sampler.unlock()
                    scrub.unlock()
                    playfield.blit(scrub, (i.posX, i.posY))
                elif i.drawButDontOverwrite:
                    blitBehind(i.frames[0], playfield, (i.posX, i.posY))
                else:
                    playfield.blit(i.frames[0], (i.posX, i.posY))
        
        # TODO: nasty hack related to the manhandling of subtracting masks
        playfield.set_colorkey(g.DEFAULT_COLOURKEY)
        playfield = playfield.convert_alpha()
        playfield.set_colorkey(g.DEFAULT_COLOURKEY2)
        playfield = playfield.convert_alpha()    

        if crop:
            mini = playfield.get_width()
            maxi = 0
            for t in self.terrains:
                if t.posX < mini:
                    mini = t.posX
                if t.posX + g.styleCache.get(t.style).terraDefs[t.itemID].sourceImg.get_width() > maxi: 
                    maxi = t.posX + g.styleCache.get(t.style).terraDefs[t.itemID].sourceImg.get_width()
            maxi = min(maxi,g.MAX_LEVEL_WIDTH-1)
            mini = max(mini,0)
            # print mini,maxi
            playfield = playfield.subsurface((mini,0,maxi-mini,playfield.get_height()))
                
        if preview:
            w = 280 
            h = 54
            prevRatio = float(w)/float(h)

            if playfield.get_width() <= prevRatio*playfield.get_height():
                playfield_t = pygame.transform.rotozoom(playfield, 0, h/float(playfield.get_height()) )
            else:
                playfield_t = pygame.transform.rotozoom(playfield, 0, w/float(playfield.get_width()) )
            playfield = pygame.Surface([playfield_t.get_width(),h+20],SWSURFACE|SRCALPHA).convert_alpha()
            playfield.fill((0,0,0,0))
            playfield.blit(playfield_t,(0,10+(h-playfield_t.get_height())/2))
        #else:
        #    playfield_t = pygame.Surface(playfield.get_size(), HWSURFACE|SRCCOLORKEY).convert()
        #    playfield_t.fill(g.DEFAULT_COLOURKEY)
        #    playfield_t.set_colorkey(g.DEFAULT_COLOURKEY)
        #    playfield_t = playfield_t.convert()
        #    playfield_t.blit(playfield, (0,0))
        #    playfield = playfield_t    

        return playfield

    def exportXML(self, filename, useGzip=False):
        if useGzip:
            f = gzip.open(filename,"w")
        else:
            f = open(filename,"w")
        xmlinfo = xml.sax.writer.XMLDoctypeInfo()
        xmlinfo.add_element_container("level")
        xmlinfo.add_element_container("skills")
        xmlinfo.add_element_container("interacts")
        xmlinfo.add_element_container("terrains")
        xmlinfo.add_element_container("steels")
        xmlout = xml.sax.writer.PrettyPrinter( f , dtdinfo=xmlinfo )
        xmlout.startDocument()
        lev = {"name":self.levelName, "author":self.author, "version":self.version,
                "releaseRate":repr(self.releaseRate), "numReleased":repr(self.numReleased), 
                "numToSave":repr(self.numToSave), "timeLimit":repr(self.timeLimit), 
                "defaultStyle":self.defaultStyle, "startPosX":repr(self.startPosX) }
        if self.keyword:   # only add the keyword attribute if the level has custom music
            lev["keyword"] = self.keyword
        xmlout.startElement("level", lev)

        skillDict = {}
        for s in self.numSkills.keys():
            skillDict[s] = repr(self.numSkills[s])
        xmlout.startElement("skills", skillDict)
        xmlout.endElement("skills")
        
        xmlout.startElement("interacts")
        for i in self.interacts:
            xmlout.startElement("interact", {"itemID":repr(i.itemID), "posX":repr(i.posX), 
                "posY":repr(i.posY), "drawFullGraphic":repr(int(i.drawFullGraphic)), 
                "drawIfTerrainUnderneath":repr(int(i.drawIfTerrainUnderneath)),
                "drawButDontOverwrite":repr(int(i.drawButDontOverwrite)),
                "drawUpsideDown":repr(int(i.drawUpsideDown)), "style":i.style })
            xmlout.endElement("interact")
        xmlout.endElement("interacts")

        xmlout.startElement("terrains")
        for t in self.terrains:
            xmlout.startElement("terrain", {"itemID":repr(t.itemID), "posX":repr(t.posX), 
                "posY":repr(t.posY), "drawFullGraphic":repr(int(t.drawFullGraphic)), 
                "drawSubtractiveTerrain":repr(int(t.drawSubtractiveTerrain)),
                "drawButDontOverwrite":repr(int(t.drawButDontOverwrite)),
                "drawUpsideDown":repr(int(t.drawUpsideDown)), "style":t.style  })
            xmlout.endElement("terrain")
        xmlout.endElement("terrains")

        xmlout.startElement("steels")
        for st in self.steels:
            xmlout.startElement("steel", {"posX":repr(st.left), "posY":repr(st.top),
                "width":repr(st.width), "height":repr(st.height) })
            xmlout.endElement("steel")
        xmlout.endElement("steels")

        xmlout.endElement("level")
        xmlout.endDocument()
        f.close()


class LevelHandler(ContentHandler):
    def __init__(self, filename):
        self.level = Level()  
        self.filename = filename
        self.isInteractsArea = False     
        self.isTerrainsArea = False     
        self.isSteelsArea = False     
        
    def startElement(self, name, attrs):
        if name == "level":
            self.level.levelName = str(attrs.get("name", self.level.levelName))
            # set levelName if it is defined in the XML file, else set it to itself (default val)
            self.level.author = str(attrs.get("author", self.level.author))
            self.level.version = str(attrs.get("version", self.level.version))
            self.level.releaseRate = int(attrs.get("releaseRate", self.level.releaseRate))
            self.level.numReleased = int(attrs.get("numReleased", self.level.numReleased))
            self.level.numToSave = int(attrs.get("numToSave", self.level.numToSave))
            self.level.timeLimit = int(attrs.get("timeLimit", self.level.timeLimit))
            self.level.defaultStyle = str(attrs.get("defaultStyle", self.level.defaultStyle))
            if self.level.defaultStyle == "self": # precache level-exclusive style
                path = os.path.split(self.filename)[0] + os.path.sep + os.path.split(self.filename)[1].split(".")[0] + "/style.xml"
                g.styleCache.cache["self"] = style.importStyleFromXML(path)
            self.level.startPosX = int(attrs.get("startPosX", self.level.startPosX))
            if attrs.has_key("keyword"):            
                self.level.keyword = str(attrs["keyword"])
        elif name == "skills":
            for k in self.level.numSkills.keys():
                self.level.numSkills[k] = int(attrs.get(k, 0))
        elif name == "interacts":
            self.isInteractsArea = True
        elif (name == "interact") & self.isInteractsArea:
            self.level.interacts.append(InteractObj())
            i = self.level.interacts[-1]
            i.posX = int(attrs.get("posX", 0))
            i.posY = int(attrs.get("posY", 0))
            i.drawFullGraphic = bool(int(attrs.get("drawFullGraphic", True)))
            i.drawIfTerrainUnderneath = bool(int(attrs.get("drawIfTerrainUnderneath", False)))
            i.drawButDontOverwrite = bool(int(attrs.get("drawButDontOverwrite", False)))
            i.drawUpsideDown = bool(int(attrs.get("drawUpsideDown", False)))
            i.itemID = int(attrs.get("itemID", 0))
            i.style = str(attrs.get("style", self.level.defaultStyle))
            currStyle = g.styleCache.get(i.style)
            try:
                i.frames = [pygame.transform.flip(x, False, i.drawUpsideDown) for x in currStyle.interDefs[i.itemID].frames]
                i.objType = currStyle.interDefs[i.itemID].interType
                i.activeRect = currStyle.interDefs[i.itemID].activeRect
            except IndexError:
                print "Fucked up: "+self.level.levelName+", "+repr(i.itemID)
        elif name == "terrains":
            self.isTerrainsArea = True
        elif (name == "terrain") & self.isTerrainsArea:
            self.level.terrains.append(TerrainObj())
            t = self.level.terrains[-1]
            t.posX = int(attrs.get("posX", 0))
            t.posY = int(attrs.get("posY", 0))
            t.drawFullGraphic = bool(int(attrs.get("drawFullGraphic", True)))
            t.drawSubtractiveTerrain = bool(int(attrs.get("drawSubtractiveTerrain", False)))
            t.drawButDontOverwrite = bool(int(attrs.get("drawButDontOverwrite", False)))
            t.drawUpsideDown = bool(int(attrs.get("drawUpsideDown", False)))
            t.itemID = int(attrs.get("itemID", 0))
            t.style = str(attrs.get("style", self.level.defaultStyle))
            # TODO: update method to use flags
            if t.style == "local":
                pass #TODO: currStyle = style.importStyleFromXML()
            else:
                currStyle = g.styleCache.get(t.style) 
            t.surface = pygame.transform.flip(currStyle.terraDefs[t.itemID].sourceImg, False, t.drawUpsideDown)
            t.subtractImg = pygame.transform.flip(currStyle.terraDefs[t.itemID].subtractImg, False, t.drawUpsideDown)
        elif name == "steels":
            self.isSteelsArea = True
        elif (name == "steel") & self.isSteelsArea:
            self.level.steels.append(Rect(0,0,16,16))
            s = self.level.steels[-1]
            s.left = int(attrs.get("posX", 0))
            s.top = int(attrs.get("posY", 0))
            s.width = int(attrs.get("width", 16))
            s.height = int(attrs.get("height", 16))

    def endElement(self, name):
        if name == "interacts":
            self.isInteractsArea = False
        elif name == "terrains":
            self.isTerrainsArea = False
        elif name == "steels":
            self.isSteelsArea = False

def importLevelFromXML(filename):
    parser = xml.sax.make_parser()
    handler = LevelHandler(filename)
    parser.setContentHandler(handler)
    if filename.split(".")[-1] == "gz":
        parser.parse(gzip.open(filename))
    else:
        parser.parse(open(filename))
    return handler.level

#def normalizeLevel(collection, diffNum, levelNum):
#    numerals = [f for f in os.listdir(getLevelPath(collection, diffNum)) if (f[-4:] == ".xml") and (f[:-4].isdigit()) ]
#    numerals = g.levelDB[collection][diffNum]
#    if numerals.has_key(levelNum):
#        return levelNum
#    else:
#        x = numerals.keys()
#        x.sort()
#        return numerals[x[0]]

def loadLevel(collection, diffNum, levelNum):
    return importLevelFromXML(getLevelPath(collection, diffNum, levelNum))
    #levelNum = normalizeLevel(collection, diffNum, levelNum)
    #lpath = getLevelPath(collection, diffNum)+repr(levelNum)+".xml"
    #if os.path.isfile(lpath):
    #    return importLevelFromXML(lpath)
    #elif os.path.isfile(lpath+".gz"):
    #    return importLevelFromXML(lpath+".gz")
    #raise ValueError, "No level file found: " + lpath

def getLevelPath(collection, diffNum=None, levelNum=None):
    lpath = (g.RESOURCES_PATH + "levels"+os.path.sep+collection+os.path.sep)
    if diffNum == None:
        return lpath
    lpath += repr(diffNum)+" "+getDifficultyString(collection, diffNum)+os.path.sep
    if levelNum == None:
        return lpath 
    else:
        if os.path.isfile(lpath+repr(levelNum)+".xml"):
            return lpath+repr(levelNum)+".xml"
        elif os.path.isfile(lpath+repr(levelNum)+".xml.gz"):
            return lpath+repr(levelNum)+".xml.gz"
        else: raise ValueError, "No level file found: " + lpath
     

def getNumLevels(collection, diffNum):
    return len(g.levelDB[collection][diffNum])
#    return len([x for x in dir(getLevelPath(collection, diffNum)) if (x.split(".")[-1].lower() == "xml")])

def getLatestDifficulty(collection): # used to autoselect in the QuickLoad dialog
    dict = g.levelDB[collection]
    diff = dict.keys(); diff.sort()
    result = diff[0]
    for d in xrange(len(diff)):
        lev = [k for k in dict[diff[d]].keys() if not k.__class__ == str]; lev.sort()
        if not dict[diff[d]][lev[-1]].access:
            result = diff[d]
            break
    return result

def getLatestLevel(collection, difficulty): # used to autoselect in the QuickLoad dialog
    dict = g.levelDB[collection][difficulty]
    lev = [k for k in dict.keys() if not k.__class__ == str]; lev.sort()
    result = lev[0]
    for l in xrange(len(lev)):
        if not dict[lev[l]].access:
            break
        result = lev[l]
    return result

def makeLevelDB():
    # this is compiled on startup and is used to store information
    # such as which levels are available, and what score was achieved
    print "Please remain absolutely calm, the game is rebuilding the level database. Only takes a few secs."
    for c in g.levelDB.keys():
        if not os.path.isdir(getLevelPath(c)):
            g.levelDB.pop(c)
            continue
        for d in g.levelDB[c].keys():
            if not os.path.isdir(getLevelPath(c,d)):
                g.levelDB[c].pop(d)
                continue
            for l in g.levelDB[c][d].keys():
                if not g.levelDB[c][d][l].__class__ == str:
                    try:
                        getLevelPath(c,d,l)
                    except ValueError:
                        g.levelDB[c][d].pop(l)
    
    cTestPath = g.RESOURCES_PATH+"levels" + os.path.sep
    collections = [n for n in os.listdir(cTestPath) if os.path.isdir(cTestPath + n)]
    collections.sort()
    for c in collections:
        if not g.levelDB.has_key(c):
            g.levelDB[c] = {}
        dTestPath = cTestPath + c + os.path.sep
        difficulties = [(int(x.split(" ")[0]), x.split(" ")[-1]) for x in os.listdir(dTestPath) if os.path.isdir(dTestPath + x)]
        difficulties.sort()
        for d in difficulties:
            if not g.levelDB[c].has_key(d[0]):
                g.levelDB[c][d[0]] = {"name": d[1]}
            lTestPath = dTestPath + repr(d[0])+" "+d[1] + os.path.sep
            # HACK: probably a better way of clarifying which files are what
            levels = [int(x.split(".")[0]) for x in os.listdir(lTestPath) if os.path.isfile(lTestPath + x) and not (x.find("xml") == -1)]
            levels.sort()
            if not g.levelDB[c][d[0]].has_key(levels[0]):
                g.levelDB[c][d[0]][levels[0]] = player.Record(True, loadLevel(c, d[0], levels[0]).levelName)
            for l in levels[1:]:
                if not g.levelDB[c][d[0]].has_key(l):
                    g.levelDB[c][d[0]][l] = player.Record(False, loadLevel(c, d[0], l).levelName)
    return
        

def getDifficultyString(collection, diffNum):
    return g.levelDB[collection][diffNum]["name"]
    #collPath = g.RESOURCES_PATH + "levels" + os.path.sep + collection + os.path.sep
    #x = [m for m in os.listdir(collPath) if os.path.isdir(collPath+m)]
    #for i in x:
    #    if i[:1]==repr(diffNum):
    #        return i[2:]
    #raise ValueError,"No such difficulty level present"

def blitBehind(source, target, targetPos, srcRect=None):
    if not srcRect:
        srcRect = source.get_rect()
    trgRect = Rect(targetPos, srcRect.size)
    temp = pygame.Surface(srcRect.size,SWSURFACE|SRCALPHA|SRCCOLORKEY).convert_alpha()
    temp.fill((0,0,0,0))
    temp.blit(source, (0,0), srcRect)
    temp.blit(target, (0,0), trgRect)
    target.fill((0,0,0,0),trgRect)
    target.blit(temp, targetPos)
    return
    
def clearArea(source, rect):
    source.fill((0,0,0,0),rect)
