"""s_blocker.py

Part of the Tundra project.
(C) 2006 Rebuild Software
http://rebui.ld.net.au

This file contains the classic definition of the lemming blocking action.
"""

import globals as g
from excepts import *

attrs = {"isDeathAction": False, "isOnGround": False, "mustStartOnGround": True, "propertyChangeOnly": False}
frames = {1:[[],None]}
# AHA! see, that's the clever thing. technically the lemming IS on the ground, and the method CHECKS
# it is still on the ground, but the THINGS YOU CAN DO are exactly the same as if you were falling
# through the air. hence, isOnGround is False.

height = 8
span = (-4,4,-3,3)
y1 = g.MAP_KEY["turnLeft"][1]
y2 = g.MAP_KEY["turnRight"][1]

def update(lemming):
    if not lemming.checkPos((lemming.posX, lemming.posY+1), "ground" ):
        cleanUp(lemming)
        lemming.setSkill("s_faller")
    else:
        lemming._incAnimFrame()
    return

def cleanUp(lemming):
    lemming.game.collisionMask.lock()    
    for i in range(-4,height):
        for dx in range(0,len(span),2):
            try:        
                p = ow,ox,oy,oz = lemming.game.collisionMask.get_at((lemming.posX+span[dx],lemming.posY-i))                
                lemming.game.collisionMask.fill( (ow,ox,oy & ~y1,oz), ((lemming.posX+span[dx],lemming.posY-i),(1,1)))
                p = ow,ox,oy,oz = lemming.game.collisionMask.get_at((lemming.posX+span[dx+1],lemming.posY-i))                
                lemming.game.collisionMask.fill( (ow,ox,oy & ~y2,oz), ((lemming.posX+span[dx+1],lemming.posY-i),(1,1)))
            except IndexError:
                continue
    lemming.game.collisionMask.unlock()
    lemming.currentSkill = "s_walker"   # HACK: workaround to make s_goner ignore the "isOnGround: False" 
    return

def set(lemming):
    lemming.currentSkill = "s_blocker"
    lemming.currentFrame = 0
    lemming.rect.left = frames[1][1].left
    lemming.rect.right = frames[1][1].right
    lemming.rect.width = frames[1][1].width
    lemming.rect.height = frames[1][1].height
    lemming.frameCount = len(frames[1][0])
    lemming.game.collisionMask.lock()    
    for i in range(-4,height):
        for dx in range(0,len(span),2):
            try:
                p = ow,ox,oy,oz = lemming.game.collisionMask.get_at((lemming.posX+span[dx],lemming.posY-i))                
                lemming.game.collisionMask.fill( (ow,ox,y1|oy,oz), ((lemming.posX+span[dx],lemming.posY-i),(1,1)))
                p = ow,ox,oy,oz = lemming.game.collisionMask.get_at((lemming.posX+span[dx+1],lemming.posY-i))                
                lemming.game.collisionMask.fill( (ow,ox,y2|oy,oz), ((lemming.posX+span[dx+1],lemming.posY-i),(1,1)))
            except IndexError:
                continue
    lemming.game.collisionMask.unlock()
    lemming._incAnimFrame()
    return
